package esmith::console::perform_backup;
use strict;
use warnings;
use esmith::ConfigDB;
use esmith::console;
use esmith::util;
use Locale::gettext;
use esmith::Backup;
use Carp;
use feature qw( say );
use esmith::BlockDevices;
use POSIX qw(:sys_wait_h strftime);
use File::stat;
use Taint::Util;

my $EMPTY = q{};

# lock file.. see bug 9127
my $backup_lock;

sub new
{
    my $class = shift;
    my $self = {
        name => gettext("Perform backup to removable media"),
        order => 80,
        };
    bless $self, $class;
    return $self;
}

sub name
{ 
    return $_[0]->{name};
}

sub order
{
    return $_[0]->{order};
}

sub backup_size
{
    my $self = shift;

}

# subs to set and remove lock on backup

sub SetLock
{
    print "Setting backup lock file\n";
    $backup_lock = esmith::Backup::set_lock;
    return $backup_lock;
}

sub RemoveLock
{
    if (defined($backup_lock))
    {
		print "Removing backup lock file\n";
	    esmith::Backup::remove_lock($backup_lock);
        $backup_lock = undef;
    }
}


sub make_backup_callback
{
    my ($device, $CompressionLevel) = @_;
    return sub {
        my $fh = shift;
        my @backup_list = esmith::Backup->restore_list;
        my @backup_excludes = esmith::Backup->excludes;
        my $backup_size = backupSize (@backup_list);
        
        # set lock.. if not, exit
		unless (SetLock()) {
			die "Error: failed to create lock file.. is a backup already running?";
		}

        open(OLDSTDOUT, ">&STDOUT");
        unless (open(STDOUT, ">$device/smeserver.tgz"))
        {
            return gettext("Could not create backup file on device").": $!\n";
        }

        open(OLDSTDERR, ">&STDERR");
        my $logger = open(STDERR, "|-");
        die "Can't fork: $!\n" unless defined $logger;

        unless ($logger)
        {
            exec qw(/usr/bin/logger -p local1.info -t console_backup);
        }

        my $status = 0;

        my $gzip = open(GZIP, "|-");
        return "could not run gzip" unless defined $gzip;        
        unless ($gzip)
        {
            close $fh;
            exec "gzip", $CompressionLevel;
        }

        my $pv = open(PV, "|-");
        return "could not run pv" unless defined $pv;
        unless ($pv)
        {
            open(STDOUT, ">&GZIP");
            close GZIP;
            open(STDERR, ">&$fh");
            exec qw(pv -i 0.2 -n -s), $backup_size
        }

        my $tar = fork;
        return "could not run tar" unless defined $tar;
        unless ($tar)
        {
            open(STDOUT, ">&PV");
            close PV;
            close GZIP;
            close $fh;
            chdir "/";
            #Create the archive
            my @directories = grep { -e $_ } @backup_list;
            my @exclude = map ("--exclude=$_",@backup_excludes);
            exec ("/bin/tar cf - @directories @exclude");
       }

        waitpid($tar, 0);
        warn "status from tar was $?\n" if $?;
        unless (close PV)
        {
            $status |= $! ? $! : $?;
            warn "status from pv is $status\n" if $status;
        }
        unless (close GZIP)
        {
            $status |= $! ? $! : $?;
            warn "status from gzip is $status\n" if $status;
        }

        open(STDOUT, ">&OLDSTDOUT");
        open(STDERR, ">&OLDSTDERR");
        close(OLDSTDERR);
        close(OLDSTDOUT);
        RemoveLock();
        return $status ? gettext("Backup failed. Look at the log files for more details.") : gettext("Backup successfully created.");
    };
}

sub doit
{
    my ($self, $console, $db) = @_;
    my @backup_list = esmith::Backup->restore_list;
    my @backup_excludes = esmith::Backup->excludes;
    my $compressionLevel = $db->get_prop('backupconsole', 'CompressionLevel') || '-6';
    my $mountpoint       = $db->get_prop('backupconsole', 'Mountpoint') || '/mnt/bootstrap-console-backup';
    my $allowMounted     = $db->get_prop('backupconsole', 'AllowMounted') || 'disabled'; ### For future use

    $ENV{PATH} = "/bin:/usr/bin";
    $ENV{HOME} = "/root";

    my $devices = esmith::BlockDevices->new ('mount' => $mountpoint, 'allowmount' => $allowMounted);
    
    INITIATE_BACKUP:
    my ($rc, $choice) = $console->yesno_page
        (
         title   => gettext("Create Backup to removable media"),
         defaultno => 1,
         text =>       gettext('Do you wish to create a backup on removable media?')."\n\n".
                      gettext('Insert removable media before proceeding.')."\n".
                      gettext('It may take many seconds to scan for media.'),
                 );
    if ($rc != 0)
    {
        $devices->destroy;
        return;
    }
    ### determine which filesystems are valid or not for backups
    # check expected backup size
    my $backup_size = backupSize (@backup_list);
    # validate each filesystem
    my ($valid, $invalid) = $devices->checkBackupDrives ($backup_size);
    my $text = $EMPTY;
    
    if (${$invalid}[0]) # If there are filesystems that are not valid.
    {
        $text .= gettext ('These filesystems are not valid:')."\n";
        foreach my $drive (sort @{$invalid})
        {
            $text .= "$drive ".$devices->desc($drive).' '.gettext ('Reason').': '.$devices->reason($drive)."\n";
        }
        $text .= "\n";
    }
    unless (${$valid}[0]) # Unless a device is found show error page
    {
        my $title = gettext('No valid backup device found').' '.gettext('size').' '.esmith::BlockDevices::scaleIt($backup_size);
        $text .= "\n$title, ".gettext('please try again');
        ($rc, $choice) = $console->yesno_page
            (
             title => $title,
             text  => $text,
             left => gettext('Try again'),
             right => gettext('Cancel'),
            ); 
        if ($rc == 0) # Try Again
        {
            goto INITIATE_BACKUP;
        }
        else
        {
            $devices->destroy;
            return;
        }
    }    
    $text .= gettext ('The following are valid for backup').' (';
    $text .= gettext ('size').' '.esmith::BlockDevices::scaleIt($backup_size).')';

    #ToDo when valid + invalid > 13 then may need to limit the information
    my @args = map { $_ => $devices->desc($_) } @{$valid};
 
    # Display the available backup destinations.
    ($rc, $choice) = $console->menu_page
        (
         title => gettext('Choose device to use for backup').' '.gettext('size').' '.esmith::BlockDevices::scaleIt($backup_size),
         text  => $text,
         argsref => \@args,
         left => gettext('Cancel'),
         right => gettext('OK'),
        );
    goto INITIATE_BACKUP unless ($rc == 0);
    untaint $choice;
    $devices->mount ($choice); # mount the chosen filesystem
    
    if  (@backup_excludes) {
        my $backupexclude = join ("\n/", sort @backup_excludes);
        ($rc, $choice) = $console->yesno_page
        (
         title   => gettext("Some parts are excluded of your backup"),
         left => gettext("Next"),
         right => gettext("Cancel"),
         text => '/' . $backupexclude,
        );
        return unless $rc == 0;
    }

    $console->infobox(
           title => gettext("Preparing for backup"),
           text => gettext("Please stand by while the system is prepared for backup..."),
          );

    system("/sbin/e-smith/signal-event", "pre-backup");

    $console->gauge(make_backup_callback($mountpoint,$compressionLevel), 'title' => gettext('Creating backup file'));

    $devices->destroy;
    
    system("/sbin/e-smith/signal-event", 'post-backup');
    RemoveLock();
    $console->message_page
        (
         title => gettext('Backup complete'),
         text  => gettext('Remove backup media.'),
        );
    return;
}


sub backupSize
{
    my $size;
    
    unless (open(DU, "-|"))
    {
        open(STDERR, ">/dev/null");
        exec qw(/usr/bin/du -sb), map { "/$_" } @_;
    }
    while (<DU>)
    {
        next unless (/^(\d+)/);
        $size += $1;
    }
    close DU;
    
    return $size;
}

#use esmith::console;
#esmith::console::perform_backup->new->doit(esmith::console->new,
# esmith::ConfigDB->open);
1;
